/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */

package fr.amap.lidar.amapvox.nephos3d.resources.postprocesses.base;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import com.jogamp.opengl.GLException;
import java.nio.IntBuffer;
import fr.amap.lidar.amapvox.nephos3d.resources.Resource;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;

/**
 * OpenGL framebuffer class.
 */
@ThreadSafe
public class FrameBuffer extends Resource {
    
    /**
     * Frame buffer object.
     */
    @GuardedBy("this")
    protected final IntBuffer fbo;
    
    /**
     * Texture buffer object.
     */
    @GuardedBy("this")
    private final IntBuffer tbo;
    
    /**
     * Render buffer object.
     */
    @GuardedBy("this")
    private final IntBuffer rbo;
    
    /**
     * The width of the texture.
     */
    protected final int width;
    
    /**
     * The height of the texture.
     */
    protected final int height;
    
    public FrameBuffer(final String identifier, final int width, 
            final int height) {
        super(identifier);
        this.fbo = Buffers.newDirectIntBuffer(1);
        this.tbo = Buffers.newDirectIntBuffer(1);
        this.rbo = Buffers.newDirectIntBuffer(1);
        this.width = width;
        this.height = height;
    }
    
    public final synchronized int getColorTexture() {
        return tbo.get(0);
    }

    @Override
    protected void onLoad(final GL3bc gl) {
        /* Generate and bind the frame buffer */
        gl.glGenFramebuffers(fbo.capacity(), fbo);
        gl.glBindFramebuffer(GL3bc.GL_FRAMEBUFFER, fbo.get(0)); 
        
        /* Generate, bind, attach and unbind the color texture */
        gl.glGenTextures(tbo.capacity(), tbo);
        gl.glBindTexture(GL3bc.GL_TEXTURE_2D, tbo.get(0));
        gl.glTexImage2D(GL3bc.GL_TEXTURE_2D, 0, GL3bc.GL_RGBA16F, width, height, 0, 
                    GL3bc.GL_RGBA, GL3bc.GL_FLOAT, null);
        gl.glTexParameteri(GL3bc.GL_TEXTURE_2D, GL3bc.GL_TEXTURE_MIN_FILTER, GL3bc.GL_NEAREST);
        gl.glTexParameteri(GL3bc.GL_TEXTURE_2D, GL3bc.GL_TEXTURE_MAG_FILTER, GL3bc.GL_NEAREST);
        gl.glFramebufferTexture2D(GL3bc.GL_FRAMEBUFFER, GL3bc.GL_COLOR_ATTACHMENT0,
                GL3bc.GL_TEXTURE_2D, tbo.get(0), 0);
        
        /* Generate, bind, attach and unbind the depth stencil render buffer */
        gl.glGenRenderbuffers(rbo.capacity(), rbo);
        gl.glBindRenderbuffer(GL3bc.GL_RENDERBUFFER, rbo.get(0)); 
        gl.glRenderbufferStorage(GL3bc.GL_RENDERBUFFER,
                GL3bc.GL_DEPTH_COMPONENT, width, height);        
        gl.glFramebufferRenderbuffer(GL3bc.GL_FRAMEBUFFER,
                GL3bc.GL_DEPTH_ATTACHMENT,
                GL3bc.GL_RENDERBUFFER, rbo.get(0));
        
        /* Check if the frame buffer is complete */
        if(gl.glCheckFramebufferStatus(GL3bc.GL_FRAMEBUFFER) != GL3bc.GL_FRAMEBUFFER_COMPLETE) {
            throw(new GLException("The framebuffer status is not complete"));
        }
        
        /* Unbind the frame buffer */
        gl.glBindFramebuffer(GL3bc.GL_FRAMEBUFFER, 0);
    }

    @Override
    protected void onBind(final GL3bc gl) {
        gl.glBindFramebuffer(GL3bc.GL_FRAMEBUFFER, fbo.get(0));
    }

    @Override
    protected void onUnbind(GL3bc gl) {
        gl.glBindFramebuffer(GL3bc.GL_FRAMEBUFFER, 0);
    }

    @Override
    protected void onUnload(GL3bc gl) {
        gl.glDeleteTextures(tbo.capacity(), tbo);
        gl.glDeleteRenderbuffers(rbo.capacity(), rbo);
        gl.glDeleteFramebuffers(fbo.capacity(), fbo);
    }

}
